char _VIDEO_MODE;

int ega4_lastoperation = -1;

/*
Prints text to the screen in text mode
Arguments: Screen row, screen column, pointer to the char string to print, 
color of text, pointer to the video memory or buffer
*/
void printtxt(unsigned char row, unsigned char col, unsigned char *str, unsigned char color, unsigned char __far *dest)
{
  unsigned char __far *destptr = (char __far*)&dest[(row*COLS+col)<<1];
  while(*str)
  {
    *destptr++=*str++;
    *destptr++=color;
  }
}
void printtxt_len(unsigned char row, unsigned char col, unsigned char *str, unsigned char color, unsigned char __far *dest, unsigned length)
{
  unsigned char __far *destptr = (char __far*)&dest[(row*COLS+col)<<1];
  while(*str && length--)
  {
    *destptr++=*str++;
    *destptr++=color;
  }
}

/*
  Puts a pixel to the screen in mode 13h
*/
void put_pixel_13h(unsigned int x, unsigned int y, unsigned char color, unsigned char __far *dest)
{
  dest[(y<<8)+(y<<6)+x] = color;
}

/*
Puts a pixel to the screen in mode 0x06
*/
void put_pixel_cga(unsigned int x, unsigned int y, unsigned char color, unsigned char __far *dest)
{
  unsigned char __far *dptr;
  unsigned char set_bit = 0x80 >> (x&7);
  if(!(y&0x01))
  {
    dptr = &dest[(y>>1)*80+(x>>3)];

    if(!color)
      *dptr &= ~set_bit;
    else
      *dptr |= set_bit;
  }
  else
  {
    dptr = &dest[0x2000+(y>>1)*80+(x>>3)];
    if(!color)
      *dptr &= ~set_bit;
    else
      *dptr |= set_bit;
  }
}

/*
Puts a pixel to the screen in mode 0x09 or 0x11
*/
void put_pixel_ega2(unsigned int x, unsigned int y, unsigned char color, unsigned char __far *dest)
{
  unsigned char __far *dptr=&dest[y*80+(x>>3)];
  unsigned char set_bit = 0x80 >> (x&7);

  if(color)
    *dptr |= set_bit;
  else
    *dptr &= ~set_bit;
}

/*
Puts a pixel to the screen in mode 0x10 or 0x12
*/
void put_pixel_ega4(unsigned int x, unsigned int y, unsigned char color, unsigned char __far *dest)
{
  unsigned char __far *dptr=&dest[y*80+(x>>3)];
  unsigned char set_bit = 0x80 >> (x&7);

  if(ega4_lastoperation)
  {
    outp(0x03CE, 0x05);
    outp(0x03CF, 0x02);

    outp(0x03CE, 0x08);
    ega4_lastoperation=0;    
  }
  outp(0x03CF, set_bit);

  inp(*dptr);  
  *dptr = color;

 
}

/*
Puts a pixel to the screen (calls one of the functions above)
*/
void put_pixel(unsigned int x, unsigned int y, unsigned char color, unsigned char __far *dest)
{
  if(x < _RES_X && y < _RES_Y)
  {
    if(_VIDEO_MODE == 0x13)
    {
      put_pixel_13h(x, y, color, dest);
      return;
    }
    if(_VIDEO_MODE == 0x06)
    {
      put_pixel_cga(x, y, color, dest);
      return;
    }
    if(_VIDEO_MODE == 0x0F || _VIDEO_MODE == 0x11)
    {
      put_pixel_ega2(x, y, color, dest);
      return;
    }
    if(_VIDEO_MODE == 0x10 || _VIDEO_MODE == 0x12)
    {
      put_pixel_ega4(x, y, color, dest);
    }
  }
}

/*
Reads a pixel from the screen in mode 0x10 or 0x12
*/
unsigned char get_pixel_ega4(unsigned int x, unsigned int y, unsigned char __far *src)
{
  unsigned char __far *dptr=&src[y*80+(x>>3)];
  unsigned char set_bit = 0x80 >> (x&7);
  unsigned char retval=0;
  unsigned int n = 0x04;

  if(!ega4_lastoperation)
  {
    outp(0x03CE, 0x04);
    ega4_lastoperation=1;    
  }

  while(n--)
  {
    outp(0x03CF, n);
    retval <<=1;        
    if((*dptr)&set_bit) retval |= 1;
  }

  return retval;
  
}

/*
Reads a pixel from the screen in mode 0x06
*/
unsigned char get_pixel_cga(unsigned int x, unsigned int y, unsigned char __far *src)
{
  unsigned char __far *dptr;
  unsigned char set_bit = 0x80 >> (x&7);
  if(!(y&0x01))
    dptr = &src[(y>>1)*80+(x>>3)];
  else
    dptr = &src[0x2000+(y>>1)*80+(x>>3)];

  if(*dptr & set_bit)
    return 1;
  return 0;
}

/*
Reads a pixel from the screen in mode 0x13
*/
unsigned char get_pixel_13h(unsigned int x, unsigned int y, unsigned char __far *src)
{
  return src[(y<<8)+(y<<6)+x];
}

/*
Reads a pixel from the screen in mode 0x0F or 0x11
*/
unsigned char get_pixel_ega2(unsigned int x, unsigned int y, unsigned char __far *src)
{
  unsigned char __far *dptr=&src[y*80+(x>>3)];
  unsigned char set_bit = 0x80 >> (x&7);
  unsigned char get_bits = 0;

  if(*dptr & set_bit)
    get_bits = 0x01;

 

  return get_bits;
}

/*
Reads a pixel from the screen
Returns: color of the pixel
*/
unsigned char get_pixel(unsigned int x, unsigned int y, unsigned char __far *src)
{
  if(_VIDEO_MODE == 0x13)
    return get_pixel_13h(x, y, src);
  if(_VIDEO_MODE == 0x06)
    return get_pixel_cga(x, y, src);
  if(_VIDEO_MODE == 0x0F || _VIDEO_MODE == 0x11)
    return get_pixel_ega2(x, y, src);
  else
  //if(_VIDEO_MODE == 0x10 || _VIDEO_MODE == 0x12)
    return get_pixel_ega4(x, y, src);
}

/*
Draws the window borders to the screen when moving or resizing the window

Phase == 0 when starting to move or resize a window
Phase == 1 when moving a window, 3 when resizing a window
Phase == 2 when done
*/
void draw_move_borders(unsigned int x, unsigned int y, unsigned int width, unsigned int height, unsigned char phase)
{
  static unsigned char *bgpixels=0;
  static unsigned int oldx, oldy;
  static unsigned int oldwidth, oldheight;
  unsigned int n;
  unsigned int m=0;
  if((phase == 2) && bgpixels)
  {
    free(bgpixels);
    bgpixels=0;
    return;
  }
  drawresize:
  if(!phase && (bgpixels = malloc((((width+2)<<1) + ((height+2)<<1)+1) * sizeof(unsigned char))))
  {
    draw:
    for(n=width+2;n--;m++)
      bgpixels[m] = get_pixel(x-1+n, y-1, VGA);
    for(n=width+2;n--;m++)
      bgpixels[m] = get_pixel(x-1+n, y+height, VGA);
    for(n=height+2;n--;m++)
      bgpixels[m] = get_pixel(x-1, y-1+n, VGA);
    for(n=height+3;n--;m++)
      bgpixels[m] = get_pixel(x+width, y-1+n, VGA);
    oldx = x;
    oldy = y;
    oldwidth = width;
    oldheight = height;
    draw_borders(x, y, width, height, 0);
    return;
  }
  if(bgpixels)
  {
    for(n=oldwidth+2;n--;m++)
      put_pixel(oldx-1+n, oldy-1, bgpixels[m], VGA);
    for(n=oldwidth+2;n--;m++)
      put_pixel(oldx-1+n, oldy+oldheight, bgpixels[m], VGA);
    for(n=oldheight+2;n--;m++)
      put_pixel(oldx-1, oldy-1+n, bgpixels[m], VGA);
    for(n=oldheight+3;n--;m++)
      put_pixel(oldx+oldwidth, oldy-1+n, bgpixels[m], VGA);
    m=0;
    if(phase==3)
    {
      free(bgpixels);
      phase=0;
      goto drawresize;
    }
    goto draw;
  }
}

/*
Fills the screen with zeros
*/
void clear_screen(unsigned char __far *dest, unsigned long cnt)
{
  while(cnt--)
  {
    *dest++ = 0;
  }
}

void clear_screen_tm(unsigned char __far *dest, unsigned cnt)
{
  while(cnt--)
  {
    *dest++ = 0;
    *dest++ = 0x07;
  }
}

/*
Changes the video mode to mode
*/
void video_mode(unsigned char mode)
{
/*  // clear text mode video screen (may be needed on some computers)
  unsigned int n = 8000;
  unsigned char __far *tvideo = 0xB8000000;
  while(n--) *tvideo++ = 0;*/
  __asm
  {
    mov ah, 0x00
    mov al, mode
    int 0x10
  }
  _RES_X=640;
  _RES_Y=200;
  VGA=(unsigned char __far *)0xA0000000L;  
  switch((_VIDEO_MODE = mode))
  {
    case 0x13:
      _RES_X=320;
      _RES_Y=200;
    return;
    case 0x06:
      VGA=(unsigned char __far *)0xB8000000L;
    return;
    case 0x0F:case 0x10:
      _RES_Y=350;
    return;
    case 0x011:case 0x12:
      _RES_Y=480;
    return;
  }
}

/*
Waits for the CRT beam retrace
*/
/*void wait_refresh(void)
{
  __asm
  {
//    push dx
//    push ax
    mov dx, 0x03da
    jump1:    
    in al, dx
    test al, 0x08
    jnz jump1
    jump2:
    in al, dx
    test al, 0x08
    jz jump2
//    pop ax
//    pop dx
  }
}*/

void draw_rectangle_safetycheck(int *x0, int *y0, unsigned int *width, unsigned int *height)
{
  if(*y0 < 0)
  {
    if(*height > abs(*y0))
    {
      *height += *y0;
      *y0 = 0;
    }
    else return;
  }
  if(*x0 < 0)
  {
    if(*width > abs(*x0))
    {
      *width += *x0;
      *x0 = 0;
    }
    else return;
  }
  
  if(*y0 > _RES_Y || *x0 > _RES_X) return;
  if(*y0+*height > _RES_Y)
    *height -= (*y0+*height) - _RES_Y;
  if(*x0+*width > _RES_X)
    *width -= (*x0+*width) - _RES_X;

}

unsigned char __far *cgapointer(int x, int y)
{
  if(!(y&0x01))
    return VGA + (y>>1)*80+(x>>3);
  else
    return VGA + 0x2000+(y>>1)*80+(x>>3);
}

void draw_rectangle_cga(int x0, int y0, unsigned int width, unsigned int height, unsigned char color, unsigned char __far *dest)
{
  unsigned int m = height;
//  unsigned int l = (((x0+width)-((x0+width)&7)) - (x0))/8;;
  unsigned int l = (((x0+width)-((x0+width)&7)) - (x0)>>3);
  unsigned int n;
  unsigned char __far *vgap;
  unsigned char set_bits;
  int sc = x0;

  if(x0&7)
  {
    set_bits = 0xFF>>(x0&7);
    if(color)
      while(m--)
      {
        *cgapointer(x0, y0++) |= set_bits;
      }
    else
      while(m--)
      {
        *cgapointer(x0, y0++) &= ~set_bits;      
      }    
    m = height;
    y0-=height;
    sc+=8;
  }

  if(color) set_bits = 0xFF;
  else set_bits = 0x00;
  while(m--)
  {
    vgap = cgapointer(sc, y0++);
    for(n=l;n--;)
      *vgap++ = set_bits;
  }

  if((x0+width)&7)
  {
    set_bits = ~(0xFF>>((x0+width)&7));
    m=height;
    y0--;
    sc = (x0+width);
    if(color)
      while(m--)
      {
        *cgapointer(sc, y0--) |= set_bits;
      }
    else
      while(m--)
      {
        *cgapointer(sc, y0--) &= ~set_bits;      
      }    
  }    
}

void draw_rectangle_ega2(int x0, int y0, unsigned int width, unsigned int height, unsigned char color, unsigned char __far *dest)
{
  unsigned int m = height;
  unsigned int l = (((x0+width)-((x0+width)&7)) - (x0))>>3;
  unsigned int n;
  unsigned char __far *vgap = dest + (y0*80) + (x0>>3);
  unsigned char set_bits;

//  draw_rectangle_safetycheck(&x0, &y0, &width, &m);



  if(x0&7)
  {
    set_bits = 0xFF>>(x0&7);
    if(color)
      while(m--)
      {
        *vgap |= set_bits;
        vgap += _RES_X>>3;
      }
    else
      while(m--)
      {
        *vgap &= ~set_bits;      
        vgap += _RES_X>>3;
      }    
    vgap = dest + (y0*80) + (x0>>3) + 1;
    m = height;
  }

  if(color) set_bits = 0xFF;
  else set_bits = 0x00;
  while(m--)
  {
    for(n=l;n--;)
      *vgap++ = set_bits;
    vgap += (_RES_X>>3) - l;
  }

  if((x0+width)&7)
  {
    set_bits = ~(0xFF>>((x0+width)&7));
    vgap = dest + (y0*80) + ((x0+width)>>3);
    m=height;
    if(color)
      while(m--)
      {
        *vgap |= set_bits;
        vgap += _RES_X>>3;
      }
    else
      while(m--)
      {
        *vgap &= ~set_bits;      
        vgap += _RES_X>>3;
      }    
  }    
  
}

/*
Draws a rectangle to the screen in modes 0x10 and 0x12
*/

void draw_rectangle_ega4(int x0, int y0, unsigned int width, unsigned int height, unsigned char color, unsigned char __far *dest)
{
//  unsigned int n = (width+((x0+7)&7))/8-1; 
  unsigned int m = height;
//  unsigned int l = width/8;
  unsigned int l = (((x0+width)-((x0+width)&7)) - (x0))>>3;
  unsigned int n;
//  unsigned int l;
//  unsigned char __far *vgap = dest + (y0)*80 + (x0+7)/8;
  unsigned char __far *vgap = dest + (y0*80) + (x0>>3);


    
  

  if(ega4_lastoperation)
  {
    outp(0x03CE, 0x05);
    outp(0x03CF, 0x02);

    outp(0x03CE, 0x08);
    ega4_lastoperation=0;    
  }

  // left side
  if(x0&7)
  {
    outp(0x03CF, 0xFF>>(x0&7));
    while(m--)
    {
      inp(*vgap);      
      *vgap = color;
      vgap += _RES_X>>3;
    }
    vgap = dest + (y0*80) + (x0>>3) + 1;
    m = height;
  }
  // center
    
  outp(0x03CF, 0xFF);
  while(m--)
  {
    for(n=l;n--;)
      *vgap++ = color;
    vgap += (_RES_X>>3) - l;
  }
  
  // right side
  if((x0+width)&7)
  {
    outp(0x03CF, ~(0xFF>>((x0+width)&7)));
    vgap = dest + (y0*80) + ((x0+width)>>3);
    for(m=height;m--;)
    {
      inp(*vgap);      
      *vgap = color;
      vgap += _RES_X>>3;
    }
  }  
  
}

void draw_rectangle_px(int x0, int y0, unsigned int width, unsigned int height, unsigned char color, unsigned char __far *dest)
{
  unsigned int n;  
  while(height--)
  {
    for(n=width;n--;)
      put_pixel(x0+n, y0+height, color, dest);
  }
}

/*
Draws a rectangle to the screen
*/
void draw_rectangle(int x0, int y0, unsigned int width, unsigned int height, unsigned char color, unsigned char __far *dest)
{

  draw_rectangle_safetycheck(&x0, &y0, &width, &height);
  if(_VIDEO_MODE == 0x10 || _VIDEO_MODE == 0x12)
    draw_rectangle_ega4(x0, y0, width, height, color, dest);
  else if(_VIDEO_MODE == 0x0F || _VIDEO_MODE == 0x11)
    draw_rectangle_ega2(x0, y0, width, height, color, dest);
  else if(_VIDEO_MODE == 0x06)
    draw_rectangle_cga(x0, y0, width, height, color, dest);  
  else
    draw_rectangle_px(x0, y0, width, height, color, dest);
/*    while(height--)
    {
      for(n=width;n--;)
        put_pixel(x0+n, y0+height, color, dest);
    }*/
}

/*
Fills the screen with argument color in modes 0x10 and 0x12
*/
void fill_screen_ega4(unsigned char color)
{
  unsigned char __far *vgap = VGA + ((_RES_X)>>3) * (_RES_Y);
  unsigned int cnt = ((_RES_X)>>3) * (_RES_Y)+1;
  
  outp(0x03CE, 0x05);
  outp(0x03CF, 0x02);

  outp(0x03CE, 0x08);
  outp(0x03CF, 0xFF);  

//  inp(*VGA);

  while(cnt--)
    *vgap-- = color;

}

/*
Fills the screen with color in modes 0x0F and 0x11
*/
void fill_screen_ega2(unsigned char color)
{
  unsigned char __far *vgap = VGA + ((_RES_X)>>3) * (_RES_Y);
  unsigned int cnt = ((_RES_X)>>3) * (_RES_Y)+1;  
  if(color)
    while(cnt--)
      *vgap-- = 0xFF;
  else
    while(cnt--)
      *vgap-- = 0x00;
}

/*
Fills the screen with color in mode 0x06
*/
void fill_screen_cga(unsigned char color)
{
  unsigned char __far *vgap = VGA + ((_RES_X)>>3) * ((_RES_Y)>>1);
  unsigned int cnt = ((_RES_X)>>3) * ((_RES_Y)>>1) +1;    
  while(cnt--)
    *vgap-- = color;
  vgap = VGA + 0x2000 + ((_RES_X)>>3) * ((_RES_Y)>>1);
  cnt = ((_RES_X)>>3) * ((_RES_Y)>>1) +1;      
//  while(vgap >= VGA + 0x2000)
  while(cnt--)
    *vgap-- = color;
}

/*
  Draws a line to the screen
*/
void draw_line(unsigned int x0, unsigned int y0, unsigned int x1, unsigned int y1, unsigned char color, unsigned char __far *dest)
{
  int dx=x1-x0;
  int dy=y1-y0;
  unsigned int n,a=0;
  unsigned int absdx,absdy;

  absdx=abs(dx);
  absdy=abs(dy);

  put_pixel(x0,y0,color,dest);
  if(absdx>absdy)
    for(n=absdx;n--;a+=absdy)
    {
      if(a>=absdx)
      {
        if(dy>0)y0++;
        else y0--;
        a-=absdx;
      }
      if(dx>0)x0++;
      else x0--;
      put_pixel(x0,y0,color,dest);      
    }
  else if(absdx<absdy)
    for(n=absdy;n--;a+=absdx)
    {
      if(a>=absdy)
      {
        if(dx>0)x0++;
        else x0--;
        a-=absdy;
      }
      if(dy>0)y0++;
      else y0--;
      put_pixel(x0,y0,color,dest);            
    }
  else
    for(n=absdx;n--;)
    {
      if(dx>0)x0++;
      else x0--;
      if(dy>0)y0++;
      else y0--;
      put_pixel(x0,y0,color,dest);
    }
}
  
/*
Updates the display memory from a buffer pointed by *VGABUFF  
*/
/*void update_screen(void)
{
  unsigned long n=64000;

  VGA+=n;
  VGABUFF+=n;

  while(n--)
    *VGA-- = *VGABUFF--;    
}*/

/*
Draws rectangular borders to the screen
*/
void draw_borders(unsigned int x, unsigned int y, unsigned int width, unsigned int height, unsigned char type)
{
  unsigned char bordercolor_top = _BORDERCOLOR_SHADOW;
  unsigned char bordercolor_bottom = _BORDERCOLOR_SHADOW;
  unsigned char bordercolor_left = _BORDERCOLOR_SHADOW;
  unsigned char bordercolor_right = _BORDERCOLOR_SHADOW;
  switch (type)
  {
    // outset
    case 0:
      bordercolor_top = _BORDERCOLOR_LIGHT;
      bordercolor_left = _BORDERCOLOR_LIGHT;
      bordercolor_right = _BORDERCOLOR_SHADOW;
      bordercolor_bottom = _BORDERCOLOR_SHADOW;
    break;
    // inset
    case 1:
      bordercolor_top = _BORDERCOLOR_SHADOW;
      bordercolor_left = _BORDERCOLOR_SHADOW;
      bordercolor_right = _BORDERCOLOR_LIGHT;
      bordercolor_bottom = _BORDERCOLOR_LIGHT;
    default:
    break;
    // normal light border
    case 2:
      bordercolor_top = _BORDERCOLOR_LIGHT;
      bordercolor_left = _BORDERCOLOR_LIGHT;
      bordercolor_right = _BORDERCOLOR_LIGHT;
      bordercolor_bottom = _BORDERCOLOR_LIGHT;      
    break;
    // normal dark border
    case 3:
      bordercolor_top = _BORDERCOLOR_SHADOW;
      bordercolor_left = _BORDERCOLOR_SHADOW;    
      bordercolor_right = _BORDERCOLOR_SHADOW;
      bordercolor_bottom = _BORDERCOLOR_SHADOW;
    break;
    // desktop bgcolor border
    case 4:
      bordercolor_top = _DESKTOP_BGCOLOR;
      bordercolor_left = _DESKTOP_BGCOLOR;    
      bordercolor_right = _DESKTOP_BGCOLOR;
      bordercolor_bottom = _DESKTOP_BGCOLOR;
    break;
  }
  // draw top border
  draw_line(x-1, y-1, x+width, y-1, bordercolor_top, VGA);
  // left border
  draw_line(x-1, y-1, x-1, y+height, bordercolor_left, VGA);
  // right border
  draw_line(x+width, y-1, x+width, y+height, bordercolor_right, VGA);
  // bottom border
  draw_line(x-1, y+height, x+width, y+height, bordercolor_bottom, VGA);
}
